/*
 * Copyright © 2021 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel22-tictactoe-core.
 *
 * ipspiel22-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel22-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel22-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel22.tictactoe.core.domain.impl;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Pattern;

import de.fhdw.gaming.core.domain.GameBuilder;
import de.fhdw.gaming.core.domain.GameBuilderFactory;
import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.Strategy;
import de.fhdw.gaming.core.ui.InputProvider;
import de.fhdw.gaming.core.ui.InputProviderException;
import de.fhdw.gaming.core.ui.type.validator.MaxValueValidator;
import de.fhdw.gaming.core.ui.type.validator.MinValueValidator;
import de.fhdw.gaming.core.ui.type.validator.PatternValidator;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeGameBuilder;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeGameBuilderFactory;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToePlayer;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToePlayerBuilder;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeStrategy;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.factory.TicTacToeDefaultStrategyFactoryProvider;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.factory.TicTacToeStrategyFactory;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.factory.TicTacToeStrategyFactoryProvider;
import de.fhdw.gaming.ipspiel22.tictactoe.core.moves.factory.TicTacToeMoveFactory;
import de.fhdw.gaming.ipspiel22.tictactoe.core.moves.impl.TicTacToeDefaultMoveFactory;

/**
 * Implements {@link TicTacToeGameBuilderFactory} by creating a Tic Tac Toe game builder.
 */
public final class TicTacToeGameBuilderFactoryImpl implements TicTacToeGameBuilderFactory {

    /**
     * The number of players.
     */
    private static final int NUMBER_OF_PLAYERS = 2;
    /**
     * Minimum number of rows (and columns) of the board.
     */
    private static final int MIN_BOARD_SIZE = 2;
    /**
     * Maximum number of rows (and columns) of the board.
     */
    private static final int MAX_BOARD_SIZE = 16;
    /**
     * Smallest allowed maximum computation time per move in seconds.
     */
    private static final int MIN_MAX_COMPUTATION_TIME_PER_MOVE = 1;
    /**
     * Largest allowed maximum computation time per move in seconds.
     */
    private static final int MAX_MAX_COMPUTATION_TIME_PER_MOVE = 3600;

    /**
     * All available Tic Tac Toe strategies.
     */
    private final Set<TicTacToeStrategy> strategies;

    /**
     * Creates an Tic Tac Toe game factory. Tic Tac Toe strategies are loaded by using the
     * {@link java.util.ServiceLoader}.
     * <p>
     * This constructor is meant to be used by the {@link java.util.ServiceLoader}.
     */
    public TicTacToeGameBuilderFactoryImpl() {
        this(new TicTacToeDefaultStrategyFactoryProvider());
    }

    /**
     * Creates an Tic Tac Toe game factory.
     *
     * @param strategyFactoryProvider The {@link TicTacToeStrategyFactoryProvider} for loading Tic Tac Toe strategies.
     */
    TicTacToeGameBuilderFactoryImpl(final TicTacToeStrategyFactoryProvider strategyFactoryProvider) {
        final TicTacToeMoveFactory moveFactory = new TicTacToeDefaultMoveFactory();

        final List<TicTacToeStrategyFactory> factories = strategyFactoryProvider.getStrategyFactories();
        this.strategies = new LinkedHashSet<>();
        for (final TicTacToeStrategyFactory factory : factories) {
            this.strategies.add(factory.create(moveFactory));
        }
    }

    @Override
    public String getName() {
        return "TicTacToe";
    }

    @Override
    public int getMinimumNumberOfPlayers() {
        return TicTacToeGameBuilderFactoryImpl.NUMBER_OF_PLAYERS;
    }

    @Override
    public int getMaximumNumberOfPlayers() {
        return TicTacToeGameBuilderFactoryImpl.NUMBER_OF_PLAYERS;
    }

    @Override
    public List<? extends Strategy<?, ?, ?>> getStrategies() {
        return new ArrayList<>(this.strategies);
    }

    @Override
    public TicTacToeGameBuilder createGameBuilder(final InputProvider inputProvider) throws GameException {
        try {
            final TicTacToeGameBuilder gameBuilder = new TicTacToeGameBuilderImpl();

            @SuppressWarnings("unchecked")
            final Map<String, Object> gameData = inputProvider
                    .needInteger(
                            TicTacToeGameBuilderFactory.PARAM_BOARD_SIZE,
                            "Number of rows (and columns)",
                            Optional.of(TicTacToeGameBuilder.DEFAULT_BOARD_SIZE),
                            new MinValueValidator<>(TicTacToeGameBuilderFactoryImpl.MIN_BOARD_SIZE),
                            new MaxValueValidator<>(TicTacToeGameBuilderFactoryImpl.MAX_BOARD_SIZE))
                    .needInteger(
                            GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE,
                            "Maximum computation time per move in seconds",
                            Optional.of(GameBuilder.DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE),
                            new MinValueValidator<>(TicTacToeGameBuilderFactoryImpl.MIN_MAX_COMPUTATION_TIME_PER_MOVE),
                            new MaxValueValidator<>(TicTacToeGameBuilderFactoryImpl.MAX_MAX_COMPUTATION_TIME_PER_MOVE))
                    .requestData("Board properties");

            gameBuilder.changeBoardSize((Integer) gameData.get(TicTacToeGameBuilderFactory.PARAM_BOARD_SIZE));
            gameBuilder.changeMaximumComputationTimePerMove(
                    (Integer) gameData.get(GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE));

            final InputProvider firstPlayerInputProvider = inputProvider.getNext(gameData);
            final Map<String, Object> firstPlayerData = this
                    .requestPlayerData(firstPlayerInputProvider, "Player 1", Optional.empty());
            final TicTacToePlayer firstPlayer = this.createPlayer(gameBuilder.createPlayerBuilder(), firstPlayerData);
            final TicTacToeStrategy firstPlayerStrategy = this.getStrategy(firstPlayerData);
            gameBuilder.addPlayer(firstPlayer, firstPlayerStrategy);

            final InputProvider secondPlayerInputProvider = firstPlayerInputProvider.getNext(firstPlayerData);
            final Map<String, Object> secondPlayerData = this.requestPlayerData(
                    secondPlayerInputProvider,
                    "Player 2",
                    Optional.of(
                            !(Boolean) firstPlayerData.get(TicTacToeGameBuilderFactory.PARAM_PLAYER_USING_CROSSES)));
            final TicTacToePlayer secondPlayer = this.createPlayer(gameBuilder.createPlayerBuilder(), secondPlayerData);
            final TicTacToeStrategy secondPlayerStrategy = this.getStrategy(secondPlayerData);
            gameBuilder.addPlayer(secondPlayer, secondPlayerStrategy);

            return gameBuilder;
        } catch (final InputProviderException e) {
            throw new GameException(String.format("Creating Tic Tac Toe game was aborted: %s", e.getMessage()), e);
        }
    }

    /**
     * Initialises an Tic Tac Toe player builder.
     *
     * @param inputProvider The input provider.
     * @param title         The title for the UI.
     * @param usingCrosses  If set, determines whether the player uses crosses or noughts. Otherwise, the setting is
     *                      initially set to crosses, and the user can change it at will.
     * @return {@code playerBuilder}.
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    @SuppressWarnings("unchecked")
    private Map<String, Object> requestPlayerData(final InputProvider inputProvider, final String title,
            final Optional<Boolean> usingCrosses) throws GameException, InputProviderException {

        inputProvider
                .needString(
                        GameBuilderFactory.PARAM_PLAYER_NAME,
                        "Name",
                        Optional.empty(),
                        new PatternValidator(Pattern.compile("\\S+(\\s+\\S+)*")))
                .needBoolean(
                        TicTacToeGameBuilderFactory.PARAM_PLAYER_USING_CROSSES,
                        "Uses crosses",
                        Optional.of(Boolean.TRUE))
                .needObject(GameBuilderFactory.PARAM_PLAYER_STRATEGY, "Strategy", Optional.empty(), this.strategies);

        if (usingCrosses.isPresent()) {
            inputProvider.fixedBoolean(TicTacToeGameBuilderFactory.PARAM_PLAYER_USING_CROSSES, usingCrosses.get());
        }

        return inputProvider.requestData(title);
    }

    /**
     * Creates an Tic Tac Toe player builder.
     *
     * @param playerBuilder The player builder.
     * @param playerData    The requested player data.
     * @return {@link TicTacToePlayer}.
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    private TicTacToePlayer createPlayer(final TicTacToePlayerBuilder playerBuilder,
            final Map<String, Object> playerData) throws GameException, InputProviderException {

        return playerBuilder.changeName((String) playerData.get(GameBuilderFactory.PARAM_PLAYER_NAME))
                .changeUsingCrosses((Boolean) playerData.get(TicTacToeGameBuilderFactory.PARAM_PLAYER_USING_CROSSES))
                .build();
    }

    /**
     * Returns an Tic Tac Toe strategy.
     *
     * @param playerData The requested player data.
     * @return The Tic Tac Toe strategy.
     */
    private TicTacToeStrategy getStrategy(final Map<String, Object> playerData) {
        return (TicTacToeStrategy) playerData.get(GameBuilderFactory.PARAM_PLAYER_STRATEGY);
    }
}
