/*
 * Copyright © 2021 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel22-tictactoe-core.
 *
 * ipspiel22-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel22-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel22-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel22.tictactoe.core.domain.impl;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import de.fhdw.gaming.core.domain.DefaultObserverFactoryProvider;
import de.fhdw.gaming.core.domain.GameBuilder;
import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.ObserverFactoryProvider;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeGame;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeGameBuilder;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToePlayer;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToePlayerBuilder;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeState;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeStrategy;
import de.fhdw.gaming.ipspiel22.tictactoe.core.moves.impl.AbstractTicTacToeMove;

/**
 * Implements {@link TicTacToeGameBuilder}.
 */
final class TicTacToeGameBuilderImpl implements TicTacToeGameBuilder {

    /**
     * The {@link ObserverFactoryProvider}.
     */
    private ObserverFactoryProvider observerFactoryProvider;
    /**
     * The player using crosses.
     */
    private Optional<TicTacToePlayer> crossesPlayer;
    /**
     * The strategy of the player using crosses.
     */
    private Optional<TicTacToeStrategy> crossesPlayerStrategy;
    /**
     * The player using noughts.
     */
    private Optional<TicTacToePlayer> noughtsPlayer;
    /**
     * The strategy of the player using noughts.
     */
    private Optional<TicTacToeStrategy> noughtsPlayerStrategy;
    /**
     * The maximum computation time per move in seconds.
     */
    private int maxComputationTimePerMove;
    /**
     * The number of rows (and columns) of the board.
     */
    private int boardSize;

    /**
     * Creates an TicTacToe game builder.
     */
    TicTacToeGameBuilderImpl() {
        this.observerFactoryProvider = new DefaultObserverFactoryProvider();
        this.crossesPlayer = Optional.empty();
        this.crossesPlayerStrategy = Optional.empty();
        this.noughtsPlayer = Optional.empty();
        this.noughtsPlayerStrategy = Optional.empty();
        this.maxComputationTimePerMove = GameBuilder.DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE;
        this.boardSize = TicTacToeGameBuilder.DEFAULT_BOARD_SIZE;
    }

    @Override
    public TicTacToePlayerBuilder createPlayerBuilder() {
        return new TicTacToePlayerBuilderImpl();
    }

    @Override
    public TicTacToeGameBuilder addPlayer(final TicTacToePlayer player, final TicTacToeStrategy strategy)
            throws GameException {

        if (player.isUsingCrosses() && this.crossesPlayer.isEmpty()) {
            this.crossesPlayer = Optional.of(Objects.requireNonNull(player, "player"));
            this.crossesPlayerStrategy = Optional.of(Objects.requireNonNull(strategy, "crossesPlayerStrategy"));
        } else if (!player.isUsingCrosses() && this.noughtsPlayer.isEmpty()) {
            this.noughtsPlayer = Optional.of(Objects.requireNonNull(player, "player"));
            this.noughtsPlayerStrategy = Optional.of(Objects.requireNonNull(strategy, "noughtsPlayerStrategy"));
        } else {
            throw new GameException(
                    String.format(
                            "Adding player %s is not allowed as a player using the same type of marks "
                                    + "has already been added.",
                            player));
        }
        return this;
    }

    @Override
    public TicTacToeGameBuilder changeMaximumComputationTimePerMove(final int newMaxComputationTimePerMove) {
        this.maxComputationTimePerMove = newMaxComputationTimePerMove;
        return this;
    }

    @Override
    public TicTacToeGameBuilder changeBoardSize(final int newBoardSize) {
        this.boardSize = newBoardSize;
        return this;
    }

    @Override
    public TicTacToeGameBuilder changeObserverFactoryProvider(
            final ObserverFactoryProvider newObserverFactoryProvider) {
        this.observerFactoryProvider = newObserverFactoryProvider;
        return this;
    }

    @Override
    public TicTacToeGame build(final int id) throws GameException, InterruptedException {
        if (!this.crossesPlayer.isPresent() || !this.noughtsPlayer.isPresent()) {
            throw new GameException("An Tic Tac Toe game needs two players.");
        }

        final TicTacToeBoardImpl board = new TicTacToeBoardImpl(this.boardSize);
        final TicTacToeState initialState = new TicTacToeStateImpl(
                board,
                this.crossesPlayer.get(),
                this.noughtsPlayer.get(),
                true);

        final Map<String, TicTacToeStrategy> strategies = new LinkedHashMap<>();
        strategies.put(initialState.getCrossesPlayer().getName(), this.crossesPlayerStrategy.orElseThrow());
        strategies.put(initialState.getNoughtsPlayer().getName(), this.noughtsPlayerStrategy.orElseThrow());
        return new TicTacToeGameImpl(
                id,
                initialState,
                strategies,
                this.maxComputationTimePerMove,
                AbstractTicTacToeMove.class::isInstance,
                this.observerFactoryProvider);
    }
}
